!correct_dnb.f90
!
!Purpose
!-------
!VIIRS DNB has a substantial bias when the instrument is sun-lit but the Earth
!is in darkness. This program aims to remove the bias.
!
!Description
!-----------
!This bias has a characteristic pattern of 16 lines (= 1 scan).
!Correct for it as follows:
!
!1) Read in a 16-point reference bias, refcorr, determined off-line
!2) For each scan line, compute the 25th percentile of the radiance.
!   Assume that this radiance is largely unaffected by city lights, aurorae, etc.
!3) For each scan, compute the normalised product of the 25th percentile
!   radiance and the reference bias. Call it p.
!4) Select the region that has scsolzen in the range 70-82 deg.
!5) Compute the correlation between p and a series of reference curves. The
!   form of the reference curve is
!       (A + B*(x-x0) + C*(x-x0)^2) * 0.5*(1-tanh((x-x0)/w)
!   where A, B, C and w are constants, and x0 is to be determined. Find the
!   value of x0 that maximizes the correlation. Add 0.5 because that tends to
!   give better results.
!6) Compute the factor by which the reference curve needs to be scaled: 
!   mean amplitude of p divided by mean amplitude of the best reference curve,
!   just taking points in the contaminated regime (82>scsolzen>scsolzen(x0)).
!7) Compute the final reference curve, pfinal, for all scans.
!8) For each scan, the radiance correction for the 16 lines is
!   -pfinal(scan)*refcorr.
!9) An optional scan-dependent correction (4064 x 16) can be added if required.
!
!Restrictions
!------------
!This software is for the day to night transition (e.g. descending node in the
!northern hemisphere). Not suitable for granule-based processing: concatenate
!the granules before use (see below).
!
!Inputs/outputs
!--------------
!Input files (all binary):
!   refcorr.dat - reference correction in the contaminated regime, 16 real values
!   dnb_mean_corr.dat - optional constant scan-dependent correction, 4064*16
!                       real values. Assumed zero if the file is not present.
!   scsolzens.dat - solar angle at the satellite, 1 value per scan (real)
!   radiances.dat - scene radiances (real)
!
!Output file (binary):
!   radiances_corr.dat - corrected radiances (binary, real numbers)
!
!Compilation
!-----------
!   gfortran -O3 -o correct_dnb.exe correct_dnb.f90 hpsort.f90
!or ifort -assume byterecl -o correct_dnb.exe correct_dnb.f90 hpsort.f90
!
!Preparing the inputs
!--------------------
!To prepare the binary input files "scsolzens.dat" and "radiances.dat", from a
!set of SDR granules ("$geofiles" and "$sdrfiles"), run the following shell
!commands:
!
!  rm -f scsolzens.dat 
!  rm -f radiances.dat
!  for file in $geofiles; do 
!    h5dump -d All_Data/VIIRS-DNB-GEO_All/SCSolarZenithAngle -b LE -o scsolzen.dat $file
!    cat scsolzen.dat >>scsolzens.dat
!    rm -f scsolzen.dat
!  done
!  for file in $sdrfiles; do
!    h5dump -d All_Data/VIIRS-DNB-SDR_All/Radiance -b LE -o rad.dat $file
!    cat rad.dat >>radiances.dat
!    rm -f rad.dat
!  done
!
!History
!-----------
!03/06/2013 v1.0. Nigel Atkinson, Met Office, Exeter, UK
!
!------------------------------------------------------------------------------
! British Crown Copyright (c) 2013, the Met Office
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without 
!  modification, are permitted provided that the following conditions are met:
!
! 1. Redistributions of source code must retain the above copyright notice, 
!  this list of conditions and the following disclaimer.
!
! 2. Redistributions in binary form must reproduce the above copyright 
!  notice, this list of conditions and the following disclaimer in the 
!  documentation and/or other materials provided with the distribution.
!
! 3. The name of the author may not be used to endorse or promote products 
!  derived from this software without specific prior written permission.
! 
! THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS'' AND ANY EXPRESS OR 
! IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
! OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
! 
! IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
! SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
! TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
! THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
! (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
! OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

program correct_dnb

IMPLICIT NONE

integer, parameter :: nspot = 4064

integer :: nlines
integer :: nscan
integer :: values(13)   !for "stat"

real :: refcorr(16)      !reference radiance correction
real :: refcorrdiff(16)  !difference from mean
real :: refcorrmean      !mean reference correction
real :: radsort(nspot)   !sorted radiances
real :: radq_select(16)  !1 scan worth of 25th percentile radiances
real :: radq_diff(16)    !difference from mean
real :: radq_select_mean !mean value
real :: norm             !normalisation
real :: corr, bestcorr   !correlations
real :: amplitude
real :: const_correction(nspot,16) !constant scan-dependent correction

real :: correlation      !function

real, allocatable :: rad(:,:)      !input radiance
real, allocatable :: radcorr(:,:)  !corrected radiance
real, allocatable :: scsolzen(:)   !input spacecraft solar zenith angles
real, allocatable :: radq(:)       !25th percentile of the radiances
real, allocatable :: p(:)          !product of radiance with reference correction
real, allocatable :: testcurve(:)  !correction function tested to find the optimum solution
real, allocatable :: bestcurve(:)  !the best solution (1 value per scan)

integer :: irec, iscan, iline, i, ifirst, ilast, nselect, bestindex, ios

!------------------------------
!Read the reference corrections
!------------------------------

open(1,file='refcorr.dat',form='unformatted',status='old',access='direct',recl=64)
read(1,rec=1)refcorr
close(1)
refcorrmean = 0.0
norm = 0.0
refcorrmean = sum(refcorr(:))/16.0
refcorrdiff(:) = refcorr(:) - refcorrmean
norm = sum(refcorrdiff(:)*refcorrdiff(:))

open(1,file='dnb_mean_corr.dat',form='unformatted',status='old',access='direct', &
   recl=nspot*4,iostat=ios)
if (ios .eq. 0) then
  do i=1,16
    read(1,rec=i)const_correction(:,i)
  enddo
  close(1)
else
  write(*,*)'Assume constant correction is zero'
  const_correction(:,:) = 0.0
endif

!------------------------------
!Get number of scan lines
!------------------------------

call stat('scsolzens.dat',values)
nscan = values(8)/4
nlines = nscan*16
write(*,*)'image size:',nspot,nlines

!------------------------------
!Allocate arrays
!------------------------------

allocate (rad(1:nspot,1:nlines))
allocate (radcorr(1:nspot,1:nlines))
allocate (scsolzen(1:nscan))
allocate (radq(1:nlines))
allocate (p(1:nscan))
allocate (bestcurve(1:nscan))

!------------------------------
!Read input data
!------------------------------

open(1,file='radiances.dat',form='unformatted',status='old',access='direct', &
  recl=nspot*4)
do irec=1,nlines
  read(1,rec=irec) rad(:,irec)
enddo
close(1)

open(1,file='scsolzens.dat',form='unformatted',status='old',access='direct', &
  recl=nscan*4)
read(1,rec=1) scsolzen
close(1)

!------------------------------
!Find the 25th percentile radiances
!------------------------------

do iline=1,nlines
  radsort(:) = rad(:,iline)
  call hpsort(nspot,radsort)
  radq(iline) = radsort(nspot/4)
  if (radq(iline).LT.-1E-9.AND.iline.NE.1) radq(iline) = radq(iline-1)
enddo

  !open(1,file='rad25.dat',form='unformatted',status='replace',access='direct', &
  !  recl=nlines*4)
  !write(1,rec=1) radq
  !close(1)

!------------------------------
!Form the product
!------------------------------

do iscan=1,nscan
  iline = (iscan-1)*16
  radq_select(:) = radq(iline+1:iline+16)
  radq_select_mean = sum(radq_select(:))/16.0
  radq_diff(:) = radq_select(:) - radq_select_mean
  p(iscan) = sum(radq_diff(:)*refcorrdiff(:))/norm
enddo

  !write(*,*) p
  !open(1,file='p.dat',form='unformatted',status='replace',access='direct', &
  !  recl=nlines*4)
  !write(1,rec=1) p
  !close(1)

!------------------------------
!Find the transition region
!------------------------------

ifirst = nscan
ilast = 1
do iscan=1,nscan
  if (scsolzen(iscan) .GT. 82.0) ifirst = iscan + 1
  if (scsolzen(iscan) .GE. 70.0) ilast = iscan
enddo
nselect = ilast - ifirst + 1

!------------------------------
!Vary x0 and find the best correction curve
!------------------------------

if (nselect .gt. 1) then

  allocate(testcurve(1:nselect))
  bestcorr = 0.0
  do i=ifirst,ilast
    call corr_curve(nselect,i-ifirst+1.0,testcurve)
    corr = correlation(nselect,p(ifirst:ilast),testcurve)
    if (corr .gt. bestcorr) then
      bestindex = i
      bestcorr = corr
    endif
  enddo
  
  write(*,*) 'bestcorr, bestindex:',bestcorr, bestindex

  call corr_curve(nscan,bestindex+0.5,bestcurve)

  amplitude = sum(p(ifirst:bestindex))/sum(bestcurve(ifirst:bestindex))

!------------------------------
!Compute the final corrected radiances
!------------------------------

  do iscan=1,nscan
    iline = (iscan-1)*16
    do i=1,16
      radcorr(:,i+iline) = rad(:,i+iline) - amplitude*bestcurve(iscan)*refcorr(i) &
                           - const_correction(:,i)
    enddo
  enddo

!------------------------------
!Create output file
!------------------------------

  open(1,file='radiances_corr.dat',form='unformatted',status='replace',&
    access='direct',recl=nspot*4)
  do irec=1,nlines
    write(1,rec=irec) radcorr(:,irec)
  enddo
  close(1)

else

  write(*,*) 'Error: unable to determine the transition region.'
  write(*,*) 'Expected solar zenith angles in transition region: 82.0 to 70.0 deg'
  write(*,'(A,F5.1,A,F5.1)') 'Actual: ',scsolzen(1),' to ',scsolzen(nscan)
  call exit(1)

endif

end program correct_dnb

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine corr_curve(n,x0,x_out)   !Empirical function to model the solar contamination
  integer, intent(IN)  :: n
  real,    intent(IN)  :: x0
  real,    intent(INOUT) :: x_out(n)
  
  real :: A, B, C, w
  integer :: i
  
  A = 1.23
  B = 0.0086
  C = 1.509E-5
  w = 3.0
  
  do i=1,n
    x_out(i) = (A + B*(i-x0) + C*(i-x0)**2) * 0.5*(1-tanh((i-x0)/w))
  enddo 
end subroutine corr_curve

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

real function correlation(n,x,y)    !Pearson product-moment correlation coefficient
  integer, intent(IN)  :: n
  real,    intent(IN)  :: x(n)
  real,    intent(IN)  :: y(n)
  real :: xmean, ymean
  
  xmean = sum(x(1:n))/n
  ymean = sum(y(1:n))/n
  
  correlation = sum((x(1:n)-xmean)*(y(1:n)-ymean)) / & 
    ( sum((x(1:n)-xmean)**2)**0.5 * sum((y(1:n)-ymean)**2)**0.5 )

end function correlation
